#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<assert.h>
#include	<getopt.h>
#include	<math.h>

#include	<ace/hash.h>
#include	<ace/tree.h>
#include	<ace/reconstruct.h>

#include	"tsdb.h"

// idea for faster comparison:
// right now we parse the derivation trees, then reprint in a format that includes just rule names, lexeme names, and bracketing
// the parsing/reprinting takes a long time
// we could instead transcode the derivation tree string directly to a format missing that data
// how: delete all numbers and quoted strings. carefully, because '+' and '-' in sign-names is significant but should be deleted in numbers

struct result
{
	int			id;
	char		*surface;
	struct tree	*derivation;
};

struct result_set
{
	char			*iid;
	char			*err;
	int				nresults;
	struct result	*results;
};

int	test_generation = 0;
int	test_scores = 0;

// field numbers
int				item_id;
int				item_input;
int				parse_id;
int				parse_run_id;
int				parse_i_id;
int				parse_readings;
int				parse_treal;
int				parse_others;
int				result_parse_id;
int				result_id;
int				result_derivation;
int				result_mrs;
int				result_surface;
int				result_tree;

void	indent(int	ind)
{
	while(ind-- > 0)printf("  ");
}

int	snprint_tree_guts(char	*ptr, int	max, struct tree	*t)
{
	int	n = 0;
	if(*t->label == '"')
	{
		return snprintf(ptr, max, "(lex)");
	}
	n += snprintf(ptr+(ptr?n:0), max-(ptr?n:0), "(%s ", t->label);
	int i;
	for(i=0;i<t->ndaughters;i++)
		n += snprint_tree_guts(ptr+(ptr?n:0), max-(ptr?n:0), t->daughters[i]);
	n += snprintf(ptr+(ptr?n:0), max-(ptr?n:0), ")");
	return n;
}

print_tree_with_edge_id(struct tree	*t, int	indent, int	*edgemap)
{
	int i;
	for(i=0;i<indent;i++)printf("  ");
	if(edgemap[t->edge_id] == 0)
		printf("%s = local #%d\n", t->label, t->edge_id);
	else
		printf("%s = local #%d = remote %d\n", t->label, t->edge_id, edgemap[t->edge_id]);
	for(i=0;i<t->ndaughters;i++)
		print_tree_with_edge_id(t->daughters[i], indent+1, edgemap);
}

void	diff_trees(struct tree	*x, struct tree	*y, int	ind)
{
	if(x->label[0]=='"' && y->label[0]=='"')
	{
		indent(ind);
		printf("EQ: %s == %s\n", x->label, y->label);
		return;
	}

	if(strcmp(x->label, y->label))
	{
		indent(ind);
		printf("NEQ: %s != %s\n", x->label, y->label);
		return;
	}
	else
	{
		indent(ind);
		printf("EQ: %s == %s\n", x->label, y->label);
	}

	if(x->ndaughters != y->ndaughters)
	{
		indent(ind);
		printf("NEQ: %d daughters != %d daughters\n", x->ndaughters, y->ndaughters);
		return;
	}

	int	i;
	for(i=0;i<x->ndaughters;i++)
		diff_trees(x->daughters[i], y->daughters[i], ind+1);
}

int	*x_to_y_edge_id;
int	*y_to_x_edge_id;

record_eq_edges(int	x_eid, int	y_eid)
{
	x_to_y_edge_id[x_eid] = y_eid;
	y_to_x_edge_id[y_eid] = x_eid;
}

int	eq_tree(struct tree	*x, struct tree	*y)
{
	if(x->label[0]=='"' && y->label[0]=='"')return 1;

	if(strcmp(x->label, y->label))
	{
		//printf("%s != %s\n", x->label, y->label);
		return 0;
	}
	if(x->ndaughters != y->ndaughters)return 0;
	int	i;
	for(i=0;i<x->ndaughters;i++)
		if(!eq_tree(x->daughters[i], y->daughters[i]))
			return 0;
	// these trees are known to be identical.
	// record that about their edge IDs
	record_eq_edges(x->edge_id, y->edge_id);
	return 1;
}

int report_missing_edges(struct tree	*t, int	*edgemap)
{
	if(t->label[0]=='"')return 0;
	int	missing = 0;
	int i;
	for(i=0;i<t->ndaughters;i++)
	{
		missing += report_missing_edges(t->daughters[i], edgemap);
		//printf("#%d daughter #%d missing = %d\n", t->edge_id, t->daughters[i]->edge_id, missing);
	}
	if(!edgemap[t->edge_id] && !missing)
	{
		printf("missing local edge #%d\n", t->edge_id);
		for(i=0;i<t->ndaughters;i++)
			printf("daughter %d local ID #%d remote ID #%d\n", i, t->daughters[i]->edge_id, (edgemap[t->daughters[i]->edge_id]>0)?edgemap[t->daughters[i]->edge_id]:-1);
		extern int print_tokens;
		print_tokens = 0;
		print_tree(t, 2);
		print_tokens = 1;
		edgemap[t->edge_id] = -1;
		missing++;
	}
	else if(edgemap[t->edge_id]<=0)missing++;
	return missing;
}

fidget(struct tree	*t)
{
	FILE	*f = fopen("/tmp/res-flat-3", "r");
	assert(f != NULL);
	char	buffer[102400];
	int n = 0;
	while(fgets(buffer, 102399, f) != NULL)
	{
		struct tree	*x = string_to_tree(buffer);
		printf("compare to res-flat-3 tree %d\n", n++);
		diff_trees(t, x, 0);
	}
	fclose(f);
}

//compare_tree_lists(char	*iid, struct tree	**x, int	nx, struct tree	**y, int	ny, int	detail)
compare_tree_lists(char	*iid, struct result	*rx, int	nx, struct result	*ry, int	ny, int	detail, char	*errx, char	*erry)
{
	int	i, j;
	int		nmx = 0, nmy = 0;
	char	foundy[ny], foundx[nx];
	bzero(foundy, sizeof(foundy));
	bzero(foundx, sizeof(foundx));

	if(detail)
	{
		for(i=0;i<nx && i<ny;i++)
		{
			printf("rank %d -> lhs #%d [%f], rhs #%d [%f]\n",
				i, rx[i].id, rx[i].derivation->score, ry[i].id, ry[i].derivation->score);
		}
	}

	int	*ids = malloc(sizeof(int) * (nx+ny));
	for(i=0;i<nx+ny;i++)ids[i] = i;
	struct hash	*hx = hash_new("x hash");
	for(i=0;i<nx;i++)
	{
		int	L = snprint_tree_guts(NULL, 0, rx[i].derivation);
		char	*str = malloc(L+10);
		snprint_tree_guts(str, L+1, rx[i].derivation);
		int	*id = hash_find(hx, str);
		if(id)
		{
			printf("LHS has duplicated result %d = %d\n", i, *id);
			if(detail)
			{
				printf("tree %d\n", *id);
				print_tree(rx[*id].derivation);
				printf("tree %d\n", i);
				print_tree(rx[i].derivation);
			}
		}
		hash_add(hx, str, &ids[i]);
	}
	struct hash	*hy = hash_new("y hash");
	for(i=0;i<ny;i++)
	{
		int	L = snprint_tree_guts(NULL, 0, ry[i].derivation);
		char	*str = malloc(L+10);
		snprint_tree_guts(str, L+1, ry[i].derivation);
		int	*id = hash_find(hy, str);
		if(id)
			printf("RHS has duplicated result %d = %d\n", i, *id);
		hash_add(hy, str, &ids[i]);
		int matched = 0;
		void visitor(int	*id)
		{
			foundx[*id] = foundy[i] = 1;
			nmx++;
			if(!matched)
			{
				matched = 1;
				nmy++;
				if(test_scores)
				{
					double	sx = rx[*id].derivation->score;
					double	sy = ry[i].derivation->score;
					if(fabs(sx-sy) > 0.01)
					{
						printf("i-id %s : matching trees lhs %d and rhs %d have different scores %f and %f\n", iid, rx[*id].id, ry[i].id, sx, sy);
						if(detail)
						{
							printf(" LHS tree:\n");print_tree(rx[*id].derivation);printf("\n");
							printf(" RHS tree:\n");print_tree(ry[i].derivation);printf("\n");
						}
					}
				}
			}
		}
		hash_visit_key(hx, str, (void(*)(void*))visitor);
		/*id = hash_find(hx, str);
		if(id)
		{
			foundx[*id] = foundy[i] = 1;
			nm++;
		}*/
		if(detail && !matched)
		{
			printf("RHS result %d had no matching result in the LHS\n", i);
			printf("tree: %s\n", str);
		}
	}
	hash_free(hx);
	hash_free(hy);
	free(ids);

	if(detail)
	{
		for(i=0;i<nx;i++)
			if(!foundx[i])
		{
			int	L = snprint_tree_guts(NULL, 0, rx[i].derivation);
			char	*str = malloc(L+10);
			snprint_tree_guts(str, L+1, rx[i].derivation);
			printf("LHS result %d had no matching result in the RHS\n", i);
			printf("tree: %s\n", str);
			free(str);
		}
		if(!x_to_y_edge_id)
			x_to_y_edge_id = malloc(sizeof(int) * 1000000);
		if(!y_to_x_edge_id)
			y_to_x_edge_id = malloc(sizeof(int) * 1000000);
		bzero(x_to_y_edge_id, sizeof(int) * 1000000);
		bzero(y_to_x_edge_id, sizeof(int) * 1000000);
		int i, j;
		for(i=0;i<nx;i++)
			for(j=0;j<ny;j++)
				eq_tree(rx[i].derivation, ry[j].derivation);
	}

	int	did_print_iid = 0;
	if(nmx !=nx || nmy != ny)
	{
		char	errors[1024] = "";
		if(*errx && *erry)sprintf(errors, "L[%s], R[%s]", errx, erry);
		else if(*errx)sprintf(errors, "L[%s]", errx);
		else if(*erry)sprintf(errors, "R[%s]", erry);
		if(nmx!=nmy)printf("item id %s:		%d  < %d=%d >  %d	%s\n", iid, nx-nmx, nmx,nmy, ny-nmy, errors);
		else printf("item id %s:		%d  < %d >  %d	%s\n", iid, nx-nmx, nmx, ny-nmy, errors);
		did_print_iid = 1;
		if(detail)
		{
			if(nx != nmx)
			{
				printf("list of missing edges from LHS:\n");
				for(i=0;i<nx;i++)
					if(!foundx[i])
					{
						//printf("missing tree from LHS:\n");
						//print_tree_with_edge_id(rx[i].derivation, 2, x_to_y_edge_id);
						//fidget(rx[i].derivation);
						report_missing_edges(rx[i].derivation, x_to_y_edge_id);
					}
			}
			if(ny != nmy)
			{
				printf("list of missing edges from RHS:\n");
				for(i=0;i<ny;i++)
					if(!foundy[i])
					{
						//printf("missing tree from RHS:\n");
						//print_tree(ry[i].derivation, 2);
						report_missing_edges(ry[i].derivation, y_to_x_edge_id);
					}
			}
		}
	}

	/*if(test_scores && nx == ny)
	{
		for(i=0;i<nx;i++)
		{
			double	sx = rx[i].derivation->score;
			double	sy = ry[i].derivation->score;
			if(fabs(sx - sy) > 0.01)
			{
				int	Lx = snprint_tree_guts(NULL, 0, rx[i].derivation);
				char	*strx = malloc(Lx+10);
				snprint_tree_guts(strx, Lx+1, rx[i].derivation);

				int	Ly = snprint_tree_guts(NULL, 0, ry[i].derivation);
				char	*stry = malloc(Ly+10);
				snprint_tree_guts(stry, Ly+1, ry[i].derivation);

				if(!did_print_iid)
				{
					printf("item id %s:\n", iid);
					did_print_iid = 1;
				}
				printf("   result id = %d, %d  ;  score = %f, %f\n", rx[i].id, ry[i].id, sx, sy);
				printf("      trees %s\n", (strcmp(strx, stry)?"differ":"match"));
				free(strx);
				free(stry);
				if(detail)
				{
					printf("   LHS tree:\n");print_tree(rx[i].derivation);printf("\n");
					printf("   RHS tree:\n");print_tree(ry[i].derivation);printf("\n");
				}
			}
		}
	}*/
}

struct mrs	*tree_to_mrs(struct tree	*t, char	*context)
{
	clear_slab();
	clear_mrs();
	struct dg	*dg = reconstruct_tree(t, NULL);
	if(dg == NULL)
	{
		fprintf(stderr, "in context %s\n", context);
		assert(dg != NULL);
	}
	//printf(" ... dg =\n");
	//print_dg(dg);
	//printf("\n");
	struct mrs	*m = extract_mrs(dg);
	assert(m != NULL);
	//printf(" ... reconstructed the MRS for it.\n");
	//print_mrs(m);
	//printf("\n");
	//fflush(stdout);
	//free_mrs_stuff(m);
	return m;
}

int	check_subsumption(struct tree	*t, struct mrs	*gold, char	*i_id)
{
	if(t && gold)
	{
		//printf(" ... found a tree for it.\n");
		struct mrs	*m = tree_to_mrs(t, i_id);
		//m = external_to_internal_mrs(m);
		//printf("mrs to test:\n");
		//print_mrs(m);printf("\n");
		int	cmp = mrs_subsumes(gold, m);
		//printf(" mrs_subsumes(gold, m) = %d\n", cmp);
		return cmp?1:0;
	}
	else fprintf(stderr, "missing tree or gold MRS\n");
	return 0;
}

int	items_with_badx = 0;

//compare_surface_lists(char	*iid, char	*i_input, char	**x, struct tree	**tx, int	nx, char	**y, struct tree	**ty, int	ny, struct tree	*gold_tree, struct mrs	*gold_mrs, int	detail)
compare_surface_lists(char	*iid, char	*i_input, struct result	*rx, int	nx, struct result	*ry, int	ny, struct tree	*gold_tree, struct mrs	*gold_mrs, int	detail)
{
	int	i, j;
	int		nbadx = 0;
	char	foundy[ny], foundx[nx];
	char	bady[ny], badx[nx];
	bzero(foundy, sizeof(foundy));
	bzero(foundx, sizeof(foundx));
	bzero(bady, sizeof(bady));
	bzero(badx, sizeof(badx));
/*	for(i=0;i<nx;i++)
		if(check_subsumption(rx[i].derivation, gold_mrs, iid) == 0)
		{
			if(detail)printf("bad LHS: %s\n", rx[i].surface);
			badx[i] = 1;
			nbadx++;
		}*/
	// ACE (RHS) doesn't record derivation trees...
	/*for(i=0;i<ny;i++)
		if(check_subsumption(ry[i].derivation, gold_mrs, iid) == 0)
		{
			printf("bad RHS: %s\n", ry[i].surface);
			bady[i] = 1;
		}*/
/*/	if(!x_to_y_edge_id)
		x_to_y_edge_id = malloc(sizeof(int) * 1000000);
	if(!y_to_x_edge_id)
		y_to_x_edge_id = malloc(sizeof(int) * 1000000);*/
	int	*ids = malloc(sizeof(int) * (nx+ny));
	for(i=0;i<nx+ny;i++)ids[i] = i;
	struct hash	*hx = hash_new("x hash");
	for(i=0;i<nx;i++)
//		if(!badx[i])
	{
		int	*id = hash_find(hx, rx[i].surface);
		if(id) {} //printf("LHS has duplicated result %d = %d\n", i, *id);
		hash_add(hx, rx[i].surface, &ids[i]);
		//printf("lhs '%s'\n", rx[i].surface);
	}
	struct hash	*hy = hash_new("y hash");
	int	nmx = 0, nmy = 0;
	for(i=0;i<ny;i++)
	{
		//printf("rhs '%s'\n", ry[i].surface);
		int	*id = hash_find(hy, ry[i].surface);
		if(id) {} //printf("RHS has duplicated result %d = %d\n", i, *id);
		hash_add(hy, ry[i].surface, &ids[i]);
		void	visitor(int	*id)
		{
			if(badx[*id])return;
			if(!foundx[*id])
			{
				foundx[*id] = 1;
				nmx++;
			}
			if(!foundy[i])
			{
				foundy[i] = 1;
				nmy++;
			}
		}
		hash_visit_key(hx, ry[i].surface, (void(*)(void*))visitor);
	}
	int	complain = 0;
	if(nmx != nmy)complain = 1;
	if(nmy != ny)complain = 1;
	if(nmx != nx - nbadx)complain = 1;
	if(nbadx)items_with_badx++;
	if(detail)complain = 1;
	if(complain)
	{
		printf("item id %s:		%d [%d]  < %d = %d >  %d\n", iid, nx-nmx-nbadx, nbadx, nmx, nmy, ny-nmy);
		printf("  %s\n", i_input);
		for(i=0;i<ny;i++)
			if(!bady[i] && !hash_find(hx, ry[i].surface))
				printf("extra RHS: %s\n", ry[i].surface);
		for(i=0;i<nx;i++)
			if(!badx[i] && !hash_find(hy, rx[i].surface))
				printf("extra LHS: %s\n", rx[i].surface);
		printf("\n");
	}
	hash_free(hx);
	hash_free(hy);
	free(ids);
}

usage(char	*prog)
{
	fprintf(stderr, "usage: %s [-g gold_profile] profile1 profile2 [detail-iid]\n", prog);
	exit(-1);
}

int	get_treal(struct relation	*parsex, char	*iid)
{
	char	**parse_tuple;
	parse_tuple = tsdb_lookup_relation_with_key(parsex, parse_i_id, iid);
	if(!parse_tuple)return -1;
	return atoi(parse_tuple[parse_treal]);
}

long long	get_others(struct relation	*parsex, char	*iid)
{
	char	**parse_tuple;
	parse_tuple = tsdb_lookup_relation_with_key(parsex, parse_i_id, iid);
	if(!parse_tuple)return -1;
	return atoll(parse_tuple[parse_others]);
}

struct result_set	get_results(struct relation	*parsex, struct relation	*resultx, char	*iid)
{
	int		nresults = 0;
	char	**results = NULL;
	struct tree	**trees = NULL;

	struct result_set	rs;
	rs.iid = strdup(iid);
	rs.nresults = 0;
	rs.results = NULL;

	void result_visitor(char	**result_tuple)
	{
		rs.results = realloc(rs.results, sizeof(struct result) * (rs.nresults+1));
		struct result	*r = &rs.results[rs.nresults++];
		char	*derivation = result_tuple[result_derivation];
		char	*surface = result_tuple[result_surface];
		r->surface = strdup(surface);
		if(derivation && *derivation)
		{
			char	*s = strdup(derivation);
			r->derivation = string_to_tree(s);
			if(!r->derivation)
			{
				fprintf(stderr, "unparseable derivation:\n%s\n", derivation);
				exit(-1);
			}
			free(s);
		}
		else r->derivation = NULL;
		char	*idstr = result_tuple[result_id];
		r->id = atoi(idstr);
	}

	char	**parse_tuple;
	parse_tuple = tsdb_lookup_relation_with_key(parsex, parse_i_id, iid);
	if(!parse_tuple)
		return rs;
	tsdb_visit_relation_with_key(resultx, result_parse_id, parse_tuple[parse_id], result_visitor);
	int	parse_error = get_field(parsex, "error", "string");
	rs.err = strdup(parse_tuple[parse_error]);

	int rcmp(const struct result	*x, const struct result	*y) { return x->id - y->id; }
	qsort(rs.results, rs.nresults, sizeof(struct result), (int (*)(const void*,const void*))rcmp);

	return rs;
}

int	common_treal_x = 0, common_treal_y = 0, common_treal_n = 0;
int	additional_treal_x = 0, additional_treal_xn = 0;
int	additional_treal_y = 0, additional_treal_yn = 0;

long long	common_ram_x = 0, common_ram_y = 0;
long long	additional_ram_x = 0;
long long	additional_ram_y = 0;

main(int	argc, char	*argv[])
{
	int	ch;
	char	*gold_profile = NULL;
	while((ch = getopt(argc, argv, "g:s")) != -1) switch(ch)
	{
		case	'g':	test_generation = 1;	gold_profile = optarg; break;
		case	's':	test_scores = 1;	break;
		default:	usage(argv[0]);
	}
	
	if(!((argc-optind) == 2 || (argc-optind) == 3))
		usage(argv[0]);

	ace_load_grammar("/home/sweaglesw/cdev/ace-regression/comparison.dat");
	extern int do_reconstruct_tokens;
	do_reconstruct_tokens = 0;	// the LKB doesn't have tokens in its trees...

	struct tsdb	*tx = load_tsdb(argv[optind+0]);
	struct tsdb	*ty = load_tsdb(argv[optind+1]);
	char	*focus_iid = NULL;
	if(argc==optind+3)focus_iid = argv[optind+2];

	if(!tsdb_compatible(tx, ty))
	{
		fprintf(stderr, "those profiles have different schemas\n");
		exit(-2);
	}

	struct tsdb	*tg = NULL;
	if(gold_profile)
	{
		tg = load_tsdb(gold_profile);
		if(!tsdb_compatible(tx, tg))
		{
			fprintf(stderr, "the gold profile has a different schema\n");
			exit(-2);
		}
	}

	struct relation	*item = get_relation(tx, "item");
	item_id = get_field(item, "i-id", "integer");
	item_input = get_field(item, "i-input", "string");
	//printf("item i-id is field %d\n", item_id);
	//printf("item i-input is field %d\n", item_input);

	printf("loaded %d items\n", item->ntuples);

	struct relation	*parsex = get_relation(tx, "parse");
	struct relation	*parsey = get_relation(ty, "parse");
	struct relation	*parseg = tg?get_relation(tg, "parse"):NULL;
	parse_id = get_field(parsex, "parse-id", "integer");
	parse_run_id = get_field(parsex, "run-id", "integer");
	parse_i_id = get_field(parsex, "i-id", "integer");
	parse_readings = get_field(parsex, "readings", "integer");
	parse_treal = get_field(parsex, "treal", "integer");
	parse_others = get_field(parsex, "others", "integer");
	//printf("parse readings is field %d\n", parse_readings);

	struct relation	*resultx = get_relation(tx, "result");
	printf("%d left results    %s\n", resultx->ntuples, argv[optind+0]);
	struct relation	*resulty = get_relation(ty, "result");
	printf("%d right results   %s\n", resulty->ntuples, argv[optind+1]);
	struct relation	*resultg = tg?get_relation(tg, "result"):NULL;
	if(tg)printf("%d gold results\n", resultg->ntuples);
	result_parse_id = get_field(resultx, "parse-id", "integer");
	result_id = get_field(resultx, "result-id", "integer");
	result_derivation = get_field(resultx, "derivation", "string");
	result_mrs = get_field(resultx, "mrs", "string");
	result_surface = get_field(resultx, "surface", "string");
	result_tree = get_field(resultx, "tree", "string");
	//printf("result parse-id is field %d\n", result_parse_id);

	struct relation	*prefg = NULL;
	int				pref_parse_id, pref_result_id;
	if(tg)
	{
		prefg = get_relation(tg, "preference");
		pref_parse_id = get_field(prefg, "parse-id", "integer");
		pref_result_id = get_field(prefg, "result-id", "integer");
	}


	// see if we can count the results for each item-id
	int	i;
	for(i=0;i<item->ntuples;i++)
	{
		char	*i_id = item->tuples[i][item_id];
		char	*i_input = item->tuples[i][item_input];

		if(focus_iid)
		{
			if(strcmp(i_id, focus_iid))continue;
			printf("item %s: %s\n", i_id, item->tuples[i][item_input]);
		}

		struct result_set	rx = get_results(parsex, resultx, i_id);
		struct result_set	ry = get_results(parsey, resulty, i_id);

		int	treal_x = get_treal(parsex, i_id);
		long long	others_x = get_others(parsex, i_id);
		int	treal_y = get_treal(parsey, i_id);
		long long	others_y = get_others(parsey, i_id);
		if(treal_x != -1 && treal_y != -1 && rx.nresults == ry.nresults)
		{
			common_treal_x += treal_x;
			common_ram_x += others_x;
			common_treal_y += treal_y;
			common_ram_y += others_y;
			common_treal_n += 1;
		}
		else
		{
			if(treal_x != -1)
			{
				additional_treal_x += treal_x;
				additional_ram_x += others_x;
				additional_treal_xn += 1;
			}
			if(treal_y != -1)
			{
				additional_treal_y += treal_y;
				additional_ram_y += others_y;
				additional_treal_yn += 1;
			}
		}

		if(!test_generation)
		{
			compare_tree_lists(i_id, rx.results, rx.nresults, ry.results, ry.nresults, focus_iid?1:0, rx.err, ry.err);

			//diff_trees(results_x[0], results_y[0], 0);
			//printf("left tree:\n"); print_tree(results_x[0]);printf("\n");
			//printf("right tree:\n"); print_tree(results_y[0]);printf("\n");
		}
		else
		{
			struct tree	*gold_tree = NULL;
			struct mrs	*gold_mrs = NULL;
			if(tg)
			{
				char	*rid = NULL;
				void gold_result_visitor(char	**result_tuple)
				{
					if(strcmp(rid, result_tuple[result_id]))
						return;	// not the preferred gold result
					char	*treestr = result_tuple[result_derivation];
					assert(treestr && *treestr);
					gold_tree = string_to_tree(strdup(treestr));
					assert(gold_tree);
					char	*mrsstr = result_tuple[result_mrs];
					assert(mrsstr && *mrsstr);
	do_reconstruct_tokens = 1;
					struct mrs	*gold_mrs2 = tree_to_mrs(gold_tree, i_id);
	do_reconstruct_tokens = 0;
					struct mrs	*gold_mrs1 = parse_mrs(strdup(mrsstr));
					//printf("stored mrs [%d vars]\n", gold_mrs1->vlist.nvars);
					//print_mrs(gold_mrs1);
					//printf("\n");
					//printf("reconstructed mrs [%d vars]\n", gold_mrs2->vlist.nvars);
					//print_mrs(gold_mrs2);
					//printf("\n");
					//printf("%d vs %d vars\n", gold_mrs1->vlist.nvars, gold_mrs2->vlist.nvars);
					if(!mrs_subsumes(gold_mrs1, gold_mrs2))
					{
						fprintf(stderr, "i-id %s: stored gold MRS does not subsume reconstructed gold MRS\n", i_id);
					}
					// no longer expect reconstructed mrs to subsume stored mrs in many cases,
					// since we now add an hcons for ltop that some processors don't
					/*if(!mrs_subsumes(gold_mrs2, gold_mrs1))
					{
						fprintf(stderr, "i-id %s: reconstructed gold MRS does not subsume stored gold MRS\n", i_id);
						printf("stored gold MRS:\n");
						print_mrs(gold_mrs1);
						printf("\n");
						printf("reconstructed gold MRS:\n");
						print_mrs(gold_mrs2);
						printf("\n");
					}*/
					gold_mrs = gold_mrs1;
					//gold_mrs = external_to_internal_mrs(gold_mrs);
					gold_mrs = copy_mrs(gold_mrs);
				}
				char	**parse_tuple = tsdb_lookup_relation_with_key(parseg, parse_i_id, i_id);
				if(!parse_tuple)
					continue;
				char	*pid = parse_tuple[parse_id];
				char	**pref_tuple = tsdb_lookup_relation_with_key(prefg, pref_parse_id, pid);
				if(!pref_tuple)
					continue;
				rid = pref_tuple[pref_result_id];
				tsdb_visit_relation_with_key(resultg, result_parse_id, pid, gold_result_visitor);
			}

			compare_surface_lists(i_id, i_input, rx.results, rx.nresults, ry.results, ry.nresults, gold_tree, gold_mrs, focus_iid?1:0);
		}
	}
	if(test_generation)
	{
		printf("on %d lhs items, ignored results where the reconstructed MRS was not subsumed by the gold MRS\n", items_with_badx);
	}
	printf("performance:               LHS time / ram     RHS time / ram  \n");
	printf("  %4d comparable items: %3.3fs / %3.3fgb %3.3fs / %3.3fgb\n", common_treal_n, (double)common_treal_x / 1000, (double)common_ram_x / (1024*1024*1024), (double)common_treal_y / 1000, (double)common_ram_y / (1024*1024*1024));
	printf("  %4d left items:       %3.3fs / %3.3fgb\n", additional_treal_xn, (double)additional_treal_x / 1000, (double)additional_ram_x / (1024*1024*1024));
	printf("  %4d right items:                            %3.3fs / %3.3fgb\n", additional_treal_yn, (double)additional_treal_y / 1000, (double)additional_ram_y / (1024*1024*1024));
	return 0;
}
